#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include "lib/Socket.h"
#include "lib/Pthread.h"
#include "Client.h"
#include "lib/Stdlib.h"
#include "lib/Queue4SendData.h"
#include "lib/Log.h"
#include "lib/Time.h"
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
/*---------------------------------------------------------------------------------------------------------------------------------------------
 *
 *                                                              Structure Definitions
 *
 *-------------------------------------------------------------------------------------------------------------------------------------------*/
enum TCPState {         /* The TCP status of a peer. */
    CONNECTED       = 1,        /* The connection is connected. */
    DISCONNECTED    = 2,        /* The connection is disconnected. */
    CONNECTING      = 3         /* It is connecting to a TCP server. */
};


struct PeerInfo {       /* The information of a peer. */
    enum TCPState state;            /* Status of TCP connection. */
    char  IPStr[ 16 ];              /* IP in string format (a.b.c.d). */
    pthread_mutex_t mutex;          /* Mutual exclusion to the client. */
    pthread_cond_t isFull;          /* Condition to the client. */
    pthread_cond_t isEmpty;         /* Condition to the client. */
    struct Queue4SendData Q;                 /* Sending queue. */
};


struct ThreadArgs {
    int pos;
};





/*---------------------------------------------------------------------------------------------------------------------------------------------
 *
 *                                                              Private Variables
 *
 *-------------------------------------------------------------------------------------------------------------------------------------------*/
static struct PeerInfo PeerList[ GROUP_MAX_NUM ];                   /* Peer information list. */
static pthread_rwlock_t rwlock = PTHREAD_RWLOCK_INITIALIZER;    /* Read/Write for @PeerList. */
#ifdef LOGGING_ENABLE
static char log[ 128 ];                 /* Buffer for logging. */
#endif
static pthread_rwlock_t tbRwlock = PTHREAD_RWLOCK_INITIALIZER;    /* Read/Write for thread buffer. */
static int threadBuf[ 1024 ];
static int threadBufPos = 0;
//static sigset_t   signal_mask;  /* signals to block         */




/*---------------------------------------------------------------------------------------------------------------------------------------------
 *
 *                                                              Private Functions
 *
 *-------------------------------------------------------------------------------------------------------------------------------------------*/
static ssize_t Readline ( int index, int fd, void *vptr, size_t maxlen ) {
    size_t nbr;   /* number of bytes readed */
    ssize_t rc;   /* return code of read() */
    nbr = 0;
    char c, *ptr;
    ptr = vptr;
    while ( nbr < maxlen ) {
        if ( ( rc = read( fd, &c, sizeof( char ) ) ) == 1 ) {
            *ptr++ = c;
//            printf( "Read: %c\n", c );
            nbr++;
            if ( c == '\n' )
                break;
        } else if( rc == 0 ) {
            *ptr = 0;
//            printf( "Number of bytes read: %d\n", nbr + 1 );
            return nbr + 1;
        } else if ( errno != EINTR )
            return -1;
    }
    *ptr = 0;
    return nbr;
}


static void* THREAD_BODY ( void* v ) {
    int sockfd;                                 /* Socket descriptor. */
    char IPStr[ 16 ];                           /* IP Address of server in string format (a.b.c.d). */
    int index = *( int* )v; /*free( v );*/          /* Position in the @PeerList. */
    int tmp;                                    /* Temporary variable. */
    struct SendDataStruct data;                /* Pointer to the data information where pop from the queue. */
    struct PeerInfo* node = &PeerList[ index ]; /* Peer information. */
    char recvline[ TCP_MAXLINE + 1 ];
    int ( *handler )( char*, char* ) = NULL;
    int n, connected = 0;

    Pthread_rwlock_rdlock( &rwlock, __FILE__, __LINE__ );
    strcpy( IPStr, PeerList[ index ].IPStr );   /* Clone IP address to local buffer. */
    Pthread_rwlock_unlock( &rwlock, __FILE__, __LINE__ );

    /* Wait until local TCP server is online. */
    while ( Client_TestConnection( &sockfd, "127.0.0.1" ) == 0 ) {
        Client_CloseConnection( sockfd );
        msleep( 10 );
    }
    Client_CloseConnection( sockfd );

    /* Assume the connection is OK by default. */
    Pthread_rwlock_wrlock( &rwlock, __FILE__, __LINE__ );
    node->state = CONNECTED;
    Pthread_rwlock_unlock( &rwlock, __FILE__, __LINE__ );

    while ( 1 ) {
        if ( connected == 0 ) {
            connected = Client_TestConnection( &sockfd, IPStr );
            fprintf( stderr, "[Client.c] Connect to %s\n", IPStr );
        }

        /* Wait until messages come. */
        Pthread_mutex_lock( &node->mutex, __FILE__, __LINE__ );
        Pthread_rwlock_rdlock( &rwlock, __FILE__, __LINE__ );
        tmp = Queue4SendData_GetCount( & ( node->Q ) );
        Pthread_rwlock_unlock( &rwlock, __FILE__, __LINE__ );
        while ( tmp == 0 ) {
            Pthread_cond_wait( &node->isEmpty, &node->mutex, __FILE__, __LINE__ );
            Pthread_rwlock_rdlock( &rwlock, __FILE__, __LINE__ );
            tmp = Queue4SendData_GetCount( & ( node->Q ) );
            Pthread_rwlock_unlock( &rwlock, __FILE__, __LINE__ );
            if ( tmp == 0 )
                continue;
            break;
        }

        Pthread_rwlock_wrlock( &rwlock, __FILE__, __LINE__ );
        data = Queue4SendData_Pop( & ( node->Q ) );
        Pthread_rwlock_unlock( &rwlock, __FILE__, __LINE__ );

        /* Let Producer continue insert items. */
        Pthread_cond_signal( &node->isFull, __FILE__, __LINE__ );
        Pthread_mutex_unlock( &node->mutex, __FILE__, __LINE__ );

        if ( connected == 0 )
            continue;

        /* Print out message to be sent. */
        n = 0;
        fprintf( stderr, "\t[XXX] Sending data [" );
        while ( data.send_buf[ n ] != '\n' ) {
            fprintf( stderr, "%c", data.send_buf[ n ] );
            n++;
        }
        fprintf( stderr, "] to %s\n", IPStr );

//            fprintf( stderr, "ERR:%d\n", sockatmark( sockfd ) );
        Writen( sockfd, data.send_buf, strlen( data.send_buf ), __FILE__, __LINE__ );
        if ( ( n = Readline( index, sockfd, recvline, TCP_MAXLINE ) ) <= 0 ) {
#ifdef LOGGING_ENABLE
            if ( n == 0 )
                sprintf( log, "%s(%d) -> Server %s terminated prematurely", __FILE__, __LINE__, IPStr );
            else
                sprintf( log, "%s(%d) -> Readline() on %s failed: %s", __FILE__, __LINE__, IPStr, strerror( errno ) );
            Log( log );
#endif
            tmp = 0;
        } else {
            printf( "\t\t -> %s RETURNS %s", IPStr, recvline );
            handler = data.handler;
            handler( recvline, IPStr );     /* Call packet handler. */
        }
    }

    Client_CloseConnection( sockfd );
    Client_Destructor( IPStr );
    return NULL;
}




/*---------------------------------------------------------------------------------------------------------------------------------------------
 *
 *                                                              Public Functions
 *
 *-------------------------------------------------------------------------------------------------------------------------------------------*/
void Client_CloseConnection ( int sockfd ) {
    Close( sockfd, __FILE__, __LINE__ );
}


void Client_Constructor ( const char* IPAddr ) {
    int status;
    pthread_t tid;      /* Thread ID. */
    pthread_attr_t rx;  /* Parameter of thread. */
    int pos = Client_GetPosition( IPAddr );;
    struct PeerInfo* node = &PeerList[ pos ];
    Pthread_rwlock_rdlock( &rwlock, __FILE__, __LINE__ );
    status = node->state;
    Pthread_rwlock_unlock( &rwlock, __FILE__, __LINE__ );
    if ( status == DISCONNECTED ) {
        Pthread_rwlock_wrlock( &rwlock, __FILE__, __LINE__ );
        strcpy( node->IPStr, IPAddr );                                      /* Clone IPAddr to the @PeerList. */
        Queue4SendData_Init( &( node->Q ), SEND_QUEUE );                                 /* Initiate the sending queue. */
        node->state = CONNECTING;                                           /* Set status to connecting. */
        Pthread_rwlock_unlock( &rwlock, __FILE__, __LINE__ );
        Pthread_attr_setdetachstate( &rx, PTHREAD_CREATE_DETACHED, __FILE__, __LINE__ );

        //xxx
        Pthread_rwlock_wrlock( &tbRwlock, __FILE__, __LINE__ );
        if ( threadBufPos >= 1024 )
            threadBufPos = 0;
        threadBuf[ threadBufPos ] = pos;
        Pthread_create( &tid, &rx, THREAD_BODY, ( void* )&( threadBuf[ threadBufPos ] ), __FILE__, __LINE__ );
        threadBufPos++;
        Pthread_rwlock_unlock( &tbRwlock, __FILE__, __LINE__ );
    }
}


void Client_Destructor ( const char* IPAddr ) {
    struct PeerInfo* node = &PeerList[ Client_GetPosition( IPAddr ) ];
    Pthread_rwlock_wrlock( &rwlock, __FILE__, __LINE__ );
    Queue4SendData_Destroy( & ( node->Q ) );                                   /* Destroy the sending queue. */
    node->state = DISCONNECTED;                                 /* Set status of TCP connection to disconnected. */
    Pthread_rwlock_unlock( &rwlock, __FILE__, __LINE__ );
}


void Client_Pend ( const char* IPAddr, struct SendDataStruct *data ) {
    struct SendDataStruct local;
    struct PeerInfo* node = &PeerList[ Client_GetPosition( IPAddr ) ];
    enum TCPState TCPStatus;
    int qSize;
//    TIMER t = time2MS();
    Pthread_rwlock_rdlock( &rwlock, __FILE__, __LINE__ );
    TCPStatus = node->state;                        /* Recent TCP status of the connection. */
    Pthread_rwlock_unlock( &rwlock, __FILE__, __LINE__ );
//    printf( "%s TCP-STATUS: %d\n", IPAddr, TCPStatus );
    if ( TCPStatus == CONNECTED ) {
        /* Blocking if the queue is full. */
        Pthread_mutex_lock( &node->mutex, __FILE__, __LINE__ );
        Pthread_rwlock_rdlock( &rwlock, __FILE__, __LINE__ );
        qSize = Queue4SendData_GetCount( & ( node->Q ) );
        Pthread_rwlock_unlock( &rwlock, __FILE__, __LINE__ );
        while ( SEND_QUEUE <= qSize ) {
            Pthread_cond_wait( &node->isFull, &node->mutex, __FILE__, __LINE__ );
            Pthread_rwlock_rdlock( &rwlock, __FILE__, __LINE__ );
            qSize = Queue4SendData_GetCount( & ( node->Q ) );
            Pthread_rwlock_unlock( &rwlock, __FILE__, __LINE__ );
            if ( SEND_QUEUE <= qSize )
                continue;   /* If the Queue is full, then blocking. */
            break;
        }

        Pthread_rwlock_wrlock( &rwlock, __FILE__, __LINE__ );
        local.handler = data->handler;             /* Clone packet handler. */
        strcpy( local.send_buf, data->send_buf );  /* Clone message to temporary array. */
        if ( Queue4SendData_Insert( & ( node->Q ), local ) == -1 ) {
#ifdef LOGGING_ENABLE
            sprintf( log, "%s(%d) -> Queue is full on client %s", __FILE__, __LINE__, node->IPStr );
            Log( log );
#endif
            exit( -1 );
        } else {
        }
        Pthread_rwlock_unlock( &rwlock, __FILE__, __LINE__ );

        /* Let the consumer read the inserted item. */
        Pthread_cond_signal( &node->isEmpty, __FILE__, __LINE__ );
        Pthread_mutex_unlock( &node->mutex, __FILE__, __LINE__ );
    } else if ( TCPStatus == DISCONNECTED ) {
        Client_Constructor( IPAddr );
    }
}


void Client_Static_Initialization () {
    int i;
    for ( i = 0 ; i < GROUP_MAX_NUM ; i++ ) {
        PeerList[ i ].state = DISCONNECTED;
        Pthread_mutex_init( &( PeerList[ i ].mutex ), NULL, __FILE__, __LINE__ );
        Pthread_cond_init( &( PeerList[ i ].isFull ), NULL, __FILE__, __LINE__ );
        Pthread_cond_init( &( PeerList[ i ].isEmpty ), NULL, __FILE__, __LINE__ );
    }
}


int Client_TestConnection ( int* sockfd, const char* IPAddr ) {
    struct sockaddr_in servaddr;
    struct timeval timeo = { 0, CONNECT_TIMEOUT * 1000 };
    socklen_t len = sizeof( timeo );
    if ( ( *sockfd = Socket( AF_INET, SOCK_STREAM, 0, __FILE__, __LINE__ ) ) < 0 )
        return 0;
    bzero( &servaddr, sizeof( servaddr ) );
    setsockopt( *sockfd, SOL_SOCKET, SO_SNDTIMEO, &timeo, len );
    servaddr.sin_family      = AF_INET;
    servaddr.sin_port        = htons( TCP_PORT );
    servaddr.sin_addr.s_addr = inet_addr( IPAddr );
    if ( Connect( *sockfd, ( struct sockaddr* )&servaddr, sizeof( servaddr ), __FILE__, __LINE__ ) == -1 ) {
        if ( errno == EINPROGRESS ) {
#ifdef LOGGING_ENABLE
            sprintf( log, "[CLIENT.C][ERR] -> %s TIMEOUT", IPAddr );
            Log( log );
#endif
        } else {
        }
        return 0;
    }
    return 1;
}


int Client_GetPosition ( const char* IPAddr ) {
    return ( ( inet_addr( IPAddr ) & 0xFF000000 ) >> 24 ) + IP_ADJUST;
}


static void kkk ( char* msg, char* src ) {
    fprintf( stderr, "RECV MSG;%s FROM %s\n", msg, src );
}


//void *signal_thread (void *arg)
//{
//    int       sig_caught;    /* signal caught       */
//    int       rc;            /* returned code       */
//
//
//    rc = sigwait (&signal_mask, &sig_caught);
//    if (rc != 0) {
//        /* handle error */
//    }
//    switch (sig_caught)
//    {
//    case SIGINT:     /* process SIGINT  */
//        fprintf( stderr, "SIGINT\n" );
//        break;
//    case SIGTERM:    /* process SIGTERM */
//        fprintf( stderr, "SIGTERM\n" );
//        break;
//    default:         /* should normally not happen */
//        fprintf (stderr, "\nUnexpected signal %d\n", sig_caught);
//        break;
//    }
//    return NULL;
//}



int main ( void ) {
//    pthread_t  sig_thr_id;      /* signal handler thread ID */
//    int        rc;              /* return code              */


//    sigemptyset (&signal_mask);
//    sigaddset (&signal_mask, SIGINT);
//    sigaddset (&signal_mask, SIGTERM);
//    rc = pthread_sigmask (SIG_BLOCK, &signal_mask, NULL);
//    if (rc != 0) {
//        /* handle error */
//        fprintf( stderr, "pthread_sigmask() err\n" );
//        exit( -1 );
//    }
//    /* any newly created threads inherit the signal mask */
//
//
//    rc = pthread_create (&sig_thr_id, NULL, signal_thread, NULL);
//    if (rc != 0) {
//        fprintf( stderr, "pthread_create() err\n" );
//        exit( -1 );
//    }

    struct SendDataStruct data;
    Client_Static_Initialization();
//    Client_Constructor( "127.0.0.1" );
    while ( 1 ) {
        strcpy( data.send_buf, "A\n" );
        data.handler = kkk;
        Client_Pend( "192.168.1.252", &data );
    }
    return 0;
}
